#!/usr/bin/env python

# Process a log file and try to determine the failure mode.
#
# This is a rewrite of the processonelog.sh script, which attempts to
# classify a log by running dozens of greps on it.
#
# Much more efficient if we just read the file once and apply the
# regexps in series, right?  
#
# Unfortunately it turns out python's regexp parser is about two
# orders of magnitude slower than grep.
#
# This script started out more elegant but devolved to become more
# hard-wired in a vain attempt to match the multiple-grep performance
# (the initial version used an arbitrarily nested list of regexps that
# combined (foo|bar) branches as in the .sh script; I unpacked it to
# two levels of nesting only and split up the branches, and also
# explicitly resorted to string matching for the non-regexps -- all of
# which improved performance but still nowhere close to grep).
#
# Various other tricks have failed to match grep performance either.
#
# Right now this file stands as an object lesson of why the constants
# hidden inside O() notation are important ;-)

import mmap, time, os, operator

import re
from sys import exit, argv
from math import ceil

class logmatch(object):
    string = ''
    regexp = None
    reason = None
    tag = None
    submatch = None # list of other logmatch objects to apply if this one passes, for implementing multi-level conditional matching

    def __init__(self, string, reason, tag, submatch):
        self.string = string
        self.regexp = re.compile(string, re.MULTILINE)
        self.reason = reason
        self.tag = tag
        self.submatch = submatch

    def match(self, line):
        answer = None
        if self.regexp.search(line):
#        if re.search(self.string, line):
            if self.reason and self.tag:
                answer = (self.reason, self.tag)
            if self.submatch:
                for i in self.submatch:
                    subanswer = i.match(line)
                    if subanswer:
                        return subanswer
        return answer

def rec(str):
    return str
    #return re.compile(str, re.MULTILINE)

def domatch(ml, text):
    global rt
    
    count = 0
    for i in ml:
        it = time.time()

        # Is the pattern a regexp?
        if i[4] and False:
            found = i[0].search(text)
        else:
            # Simple string
            found = text.find(i[0]) != -1
#                found = i[0] in text # hella slow for mmap !?!
        if found:
            # Check second level conditional match strings
            if i[3] and False:
                for j in i[3]:
                    if j[4]:
                        found = j[0].search(text)
                    else:
                        found = text.find(j[0]) != -1
                    if found:
                        ot = time.time()
                        rt[count] += ot - it
                        return (j[1], j[2])
                ot = time.time()
                rt[count] += ot - it
                return (i[1], i[2])

        ot = time.time()
        rt[count] += ot - it
        count += 1
    return None
                
def initmatches():

# INSTRUCTIONS FOR KEEPING THIS TABLE FAST
#
# * Don't be lazy!  Don't use a regexp because you can't be bothered
#   getting the error formatting precise
# * Don't use a regexp when a simple string match will do
# * Be as specific as you can
# * Anchor the regexp with ^ and $ where possible.
# * Test your regexps!  A lot of these were broken or needlessly
#   general because special characters like ( and . were not escaped.
# * python's regexp matcher doesn't seem to like big (foo|bar|blee)
#   statements, so don't aggregate unrelated strings this way just
#   because they share a reason/tag
# * It is still much slower than grep!?!

    ml = [(rec('Error: mtree file ./etc/mtree/BSD.local.dist. is missing'), 'mtree', 'mtree', None, True),
          ('error in pkg_delete', 'mtree', 'mtree', None, False),
          ('filesystem was touched prior to .make install', 'mtree', 'mtree', None, False),
          ('list of extra files and directories', 'mtree', 'mtree', None, False),
          ('list of files present before this port was installed', 'mtree', 'mtree', None, False),
          ('list of filesystem changes from before and after', 'mtree', 'mtree', None, False),

          (rec('Configuration .* not supported'), 'arch', 'arch', None, True),

          (rec('(configure: error:|Script.*configure.*failed unexpectedly|script.*failed: here are the contents of)'), 'configure_error', 'configure', 
           [(rec('configure: error: cpu .* not supported'), 'arch', 'arch', None, True),
            (rec('configure: error: [Pp]erl (5.* required|version too old)'), 'perl', 'perl', None, True)]
           , True),

          ('Couldn\'t fetch it - please try', 'fetch', 'fetch', None, False),

          (rec('\.(c|cc|cxx|cpp|h|y)[1-9:]+ .+\.h: No such file'), 'missing_header', 'header',
           [(rec('(X11/.*|Xosdefs)\.h: No such file'), 'USE_XLIB', 'usexlib', None, True)]
           , True),

          (rec('Error: shared library ".*" does not exist'), 'LIB_DEPENDS', 'libdepends', None, True),

          ('parse error', 'compiler_error', 'cc', None, False),
          (rec('too (many|few) arguments to'), 'compiler_error', 'cc', None, True),
          (rec('argument.*doesn.*prototype'), 'compiler_error', 'cc', None, True),
          ('incompatible type for argument', 'compiler_error', 'cc', None, False),
          ('conflicting types for', 'compiler_error', 'cc', None, False),
          (rec('undeclared \(first use (in |)this function\)'), 'compiler_error', 'cc', None, True),
          ('incorrect number of parameters', 'compiler_error', 'cc', None, False),
          ('has incomplete type and cannot be initialized', 'compiler_error', 'cc', None, False),
          (rec('error: storage size.* isn.t known'), 'compiler_error', 'cc', None, True),
          
          ('ANSI C++ forbids', 'new_compiler_error', 'newgcc', None, False),
          ('is a contravariance violation', 'new_compiler_error', 'newgcc', None, False),
          (rec('changed for new ANSI .for. scoping'), 'new_compiler_error', 'newgcc', None, True),
          (rec('[0-9]: passing .* changes signedness'), 'new_compiler_error', 'newgcc', None, True),
          ('lacks a cast', 'new_compiler_error', 'newgcc', None, False),
          ('redeclared as different kind of symbol', 'new_compiler_error', 'newgcc', None, False),
          (rec('invalid type .* for default argument to'), 'new_compiler_error', 'newgcc', None, True),
          ('wrong type argument to unary exclamation mark', 'new_compiler_error', 'newgcc', None, False),
          ('duplicate explicit instantiation of', 'new_compiler_error', 'newgcc', None, False),
          ('incompatible types in assignment', 'new_compiler_error', 'newgcc', None, False),
          (rec('assuming . on overloaded member function'), 'new_compiler_error', 'newgcc', None, True),
          (rec('call of overloaded .* is ambiguous'), 'new_compiler_error', 'newgcc', None, True),
          (rec('declaration of C function .* conflicts with'), 'new_compiler_error', 'newgcc', None, True),
          ('initialization of non-const reference type', 'new_compiler_error', 'newgcc', None, False),
          (rec('using typedef-name .* after'), 'new_compiler_error', 'newgcc', None, True),
          (rec('[0-9]: size of array .* is too large'), 'new_compiler_error', 'newgcc', None, True),
          (rec('fixed or forbidden register .* for class'), 'new_compiler_error', 'newgcc', None, True),
          ('assignment of read-only variable', 'new_compiler_error', 'newgcc', None, False),
          ('error: label at end of compound statement', 'new_compiler_error', 'newgcc', None, False),
          (rec('error:.*(has no|is not a) member'), 'new_compiler_error', 'newgcc', None, True),
          (rec('error:.*is (private|protected)'), 'new_compiler_error', 'newgcc', None, True),
          ('error: uninitialized member', 'new_compiler_error', 'newgcc', None, False),
          ('error: unrecognized command line option', 'new_compiler_error', 'newgcc', None, False),

          ('syntax error before', 'bad_c++_code', 'badc++', None, False),
          ('friend declaration', 'bad_c++_code', 'badc++', None, False),
          ('no matching function for call to', 'bad_c++_code', 'badc++', None, False),
          ('\'main\' must return \'int\'', 'bad_c++_code', 'badc++', None, False),
          ('invalid conversion from', 'bad_c++_code', 'badc++', None, False),
          ('cannot be used as a macro name as it is an operator in C++', 'bad_c++_code', 'badc++', None, False),
          ('is not a member of type', 'bad_c++_code', 'badc++', None, False),
          ('after previous specification in', 'bad_c++_code', 'badc++', None, False),
          ('no class template named', 'bad_c++_code', 'badc++', None, False),
          ('because worst conversion for the former', 'bad_c++_code', 'badc++', None, False),
          ('better than worst conversion', 'bad_c++_code', 'badc++', None, False),
          (rec('no match for.*operator'), 'bad_c++_code', 'badc++', None, True),
          ('no match for call to', 'bad_c++_code', 'badc++', None, False),
          ('undeclared in namespace', 'bad_c++_code', 'badc++', None, False),
          ('is used as a type, but is not', 'bad_c++_code', 'badc++', None, False),
          ('error: array bound forbidden', 'bad_c++_code', 'badc++', None, False),
          ('error: class definition', 'bad_c++_code', 'badc++', None, False),
          ('error: expected constructor', 'bad_c++_code', 'badc++', None, False),
          ('error: there are no arguments', 'bad_c++_code', 'badc++', None, False),
          (rec('error:.*cast.*loses precision'), 'bad_c++_code', 'badc++', None, True),
          ('ISO C++ does not support', 'bad_C++_code', 'badc++', None, False),

          ('error: array type has incomplete element type', 'gcc4_error', 'gcc4', None, False),
          ('error: conflicts with new declaration', 'gcc4_error', 'gcc4', None, False),
          (rec('error: expected.*before .class'), 'gcc4_error', 'gcc4', None, True),
          ('error: expected primary expression', 'gcc4_error', 'gcc4', None, False),
          (rec('error: extra qualification .* on member'), 'gcc4_error', 'gcc4', None, True),
          (rec('error: .*has incomplete type'), 'gcc4_error', 'gcc4', None, True),
          (rec('error: invalid cast from type .* to type'), 'gcc4_error', 'gcc4', None, True),
          (rec('error: invalid lvalue in (assignment|decremant|increment|unary)'), 'gcc4_error', 'gcc4', None, True),
          ('error: invalid storage class for function', 'gcc4_error', 'gcc4', None, False),
          (rec('error: lvalue required as (increment operator|left operand)'), 'gcc4_error', 'gcc4', None, True),
          (rec('error: .*should have been declared inside'), 'gcc4_error', 'gcc4', None, True),
          (rec('error: static declaration of.*follows non-static declaration'), 'gcc4_error', 'gcc4', None, True),
          ('error: two or more data types in declaration specifiers', 'gcc4_error', 'gcc4', None, False),
          (rec('error: .* was not declared in this scope'), 'gcc4_error', 'gcc4', None, True),

          ('/usr/libexec/elf/ld: cannot find', 'linker_error', 'ld', None, False),
          ('undefined reference to', 'linker_error', 'ld', None, False),
          (rec('cannot open -l.*: No such file'), 'linker_error', 'ld', None, True),

          (rec('install: .*: No such file'), 'install_error', 'install', None, True),

          ('conflicts with installed package', 'depend_object', 'depend', None, False),
          ('is already installed - perhaps an older version', 'depend_object', 'depend', None, False),
          (rec('You may wish to ..make deinstall.. and install this port again'), 'depend_object', 'depend', None, True),

          ('core dumped', 'coredump', 'coredump', None, False),
# Here are the less common ones:

          ('.s: Assembler messages:', 'arch', 'arch', None, False),
          (rec('Cannot (determine .* target|find the byte order) for this architecture'), 'arch', 'arch', None, True),
          (rec('^cc1: bad value.*for -mcpu.*switch'), 'arch', 'arch', None, True),
          ('could not read symbols: File in wrong format', 'arch', 'arch', None, False),
          (rec('[Ee]rror: [Uu]nknown opcode'), 'arch', 'arch', None, True),
          (rec('error.*Unsupported architecture'), 'arch', 'arch', None, True),
          ('ENDIAN must be defined 0 or 1', 'arch', 'arch', None, False),
          ('failed to merge target-specific data', 'arch', 'arch', None, False),
          (rec('(file not recognized|failed to set dynamic section sizes): File format not recognized'), 'arch', 'arch', None, True),
          ('impossible register constraint', 'arch', 'arch', None, False),
          ('inconsistent operand constraints in an .asm', 'arch', 'arch', None, False),
          (rec('Invalid configuration.*unknown.*machine.*unknown not recognized'), 'arch', 'arch', None, True),
          ('invalid lvalue in asm statement', 'arch', 'arch', None, False),
          (rec('is only for.*, and you are running'), 'arch', 'arch', None, True),
          ('not a valid 64 bit base/index expression', 'arch', 'arch', None, False),
          (rec('relocation R_X86_64_32.*can not be used when making a shared object'), 'arch', 'arch', None, True),
          ('relocation truncated to fit: ', 'arch', 'arch', None, False),
          ('shminit failed: Function not implemented', 'arch', 'arch', None, False),
          (rec('The target cpu, .*, is not currently supported\.'), 'arch', 'arch', None, True),
          ('This architecture seems to be neither big endian nor little endian', 'arch', 'arch', None, False),
          ('unknown register name', 'arch', 'arch', None, False),
          ('Unable to correct byte order', 'arch', 'arch', None, False),
          ('Unsupported platform, sorry', 'arch', 'arch', None, False),
          ('won\'t run on this architecture', 'arch', 'arch', None, False),

          
          (rec('autoconf([0-9\-\.]*): (not found|No such file or directory)'), 'autoconf', 'autoconf', None, True),
          ('autoheader: not found', "autoheader", "autoheader", None, False),
          (rec('automake(.*): not found'), "automake", "automake", None, True),

          ('Checksum mismatch', "checksum", "checksum", None, False),
          (rec('chown:.*[Ii]nvalid argument'), "chown", "chown", None, True),
          ("Shared object \"libc.so.6\" not found, required by", "compat6x", "compat6x", None, False),
          (rec("error in dependency .*, exiting"), "depend_package", "dependpkg", None, True),
          (rec("pkg_(add|create):.*(can't find enough temporary space|projected size of .* exceeds available free space)"), "disk_full", "df", None, True),

          (rec("(Can't|unable to) open display"), 'x_windows', 'xwindows', None, True),
          ('Cannot open /dev/tty for read', 'x_windows', 'xwindows', None, False),
          ('RuntimeError: cannot open display', 'x_windows', 'xwindows', None, False),
          ('You must run this program under the X-Window System', "x_windows", "xwindows", None, False),

          ('No checksum recorded for', 'distinfo', 'distinfo', None, False),
          (rec('Either .* is out of date, or'), "distinfo_update", "distinfo", None, True),

          ("Member name contains ...", "fetch", "fetch", None, False),

          ("pnohang: killing make checksum", "fetch_timeout", "fetch-timeout", None, False),
          ("fetch: transfer timed out", "fetch_timeout", "fetch-timeout", None, False),

          ("f77: not found", "f77", "f77", None, False),
          ("f77: No such file or directory", "f77", "f77", None, False),
          ("Unable to find a fortran compiler", "f77", "f77", None, False),

          ("See <URL:http://gcc.gnu.org/bugs.html> for instructions.", "gcc_bug", "gcc-bug", None, False),

          ("Run-time system build failed for some reason", "install_error", "install", None, False),
          ("tar: Error opening archive: Failed to open.*No such file or directory", "install_error", "install", None, False),

          ("/usr/bin/ld: cannot find -l","linker_error", "ld", None, False),
          (rec("cc: .*libintl.*: No such file or directory"), 'linker_error', 'ld', None, True),
          ('cc: ndbm.so: No such file or directory', 'linker_error', 'ld', None, False),
          ('error: The X11 shared library could not be loaded', 'linker_error', 'ld', None, False),
          ('libtool: link: cannot find the library', 'linker_error', 'ld', None, False),
          ('relocation against dynamic symbol', 'linker_error', 'ld', None, False),
          (rec('Shared object.*not found, required by'), "linker_error", "ld", None, True),

          ("Could not create Makefile", "makefile", "makefile", None, False),

# XXX
#elif bzgrep -v "regression-test.continuing" $1 | grep -qE "make.*(cannot open [Mm]akefile|don.t know how to make|fatal errors encountered|No rule to make target|built-in)"; then, "makefile", "makefile", None, False),

          (rec("/usr/.*/man/.*: No such file or directory"), "manpage", "manpage", None, True),
          (rec("out of .* hunks .*--saving rejects to"), "patch", "patch", None, True),

          (rec('/usr/local/bin/perl.*:.*(not found|No such file or directory)'), 'perl', 'perl', None, True),
          (rec('cp:.*site_perl: No such file or directory'), 'perl', 'perl', None, True),
          (rec('perl(.*): Perl is not installed, try .pkg_add -r perl'), 'perl', 'perl', None, True),
          (rec('Perl .* required--this is only version'), "perl", "perl", None, True),

          ('BEGIN failed--compilation aborted at ..Makefile.PL line', "perl5", "perl5", None, False),

          ('Abort trap', 'process_failed', 'process', None, False),
          ('Bus error', 'process_failed', 'process', None, False),
          (rec('Signal 1[01]'), "process_failed", "process", None, True),

          ("python: not found", "python", "python", None, False),
          (rec("USER   PID  PPID  PGID.*JOBC STAT  TT       TIME COMMAND"), "runaway_process", "runaway", None, True),
          ("pnohang: killing make package", "runaway_process", "runaway", None, False),

          ("Segmentation fault", "segfault", "segfault", None, False),
          ("initializer element is not constant", "stdio", "stdio", None, False),
          ("structure has no member named", "struct_changes", "struct", None, False),
          ("shminit failed: Permission denied", "sysvipc", "sysvipc", None, False),

          (rec('/usr/bin/ld: cannot find -l(pthread|XThrStub)'), 'threads', 'threads', None, True),
          ('cannot find -lc_r', 'threads', 'threads', None, False),
          (rec('checking for.*lc_r\.\.\. no'), 'threads', 'threads', None, True),
          ('Error: pthreads are required to build this package', 'threads', 'threads', None, False),
          ('Please install/update your POSIX threads (pthreads) library', 'threads', 'threads', None, False),
          (rec('requires.*thread support'), 'threads', 'threads', None, True),
          (': The -pthread option is deprecated', "threads", "threads", None, False),

          ("<varargs.h> is obsolete with this version of GCC", "varargs", "varargs", None, False),
          ('read-only file system', "WRKDIR", "wrkdir", None, False), #XXX previously bzgrep -i: necessary?  Could add a re flag

# Although these can be fairly common, and thus in one sense ought to be
# earlier in the evaluation, in practice they are most often secondary
# types of errors, and thus need to be evaluated after all the specific
# cases.

          (rec('cc1.*warnings being treated as errors'), "compiler_error", "cc", None, True),
          ('tar: Error exit delayed from previous errors', "install_error", "install", None, False),
          ("Cannot stat: ", "configure_error", "configure", None, False),
          ("cd: can't cd to", "NFS", "nfs", None, False),
          ("pkg_create: make_dist: tar command failed with code", "PLIST", "plist", None, False)]

    return ml

def processfile(filename, indexfile, indexdata, errordir, ml):

    global rt;

    splitname = os.path.splitext(filename)
    bzipped = 0
    if splitname[1] == ".bz2":
        bzipped = 1
        splitname = os.path.splitext(splitname[0])

    if splitname[1] != ".log":
        print "%s is not a .log or .log.bz2 file" % filename
        exit(1)

    portname = splitname[0]

    try:
        if bzipped:
            fh = bz2.BZ2File(filename, "r")
        else:
            fh = file(filename, "r")
            mm = mmap.mmap(fh.fileno(), 0, mmap.MAP_SHARED, mmap.PROT_READ)
    except IOError:
        print "%s does not exist" % filename
        exit(1)

    # size of log file in KB, rounded up
    logsize = int(ceil(os.path.getsize(filename)/1024))

    #print "count %s " % time.time()
    # Count how many ports list this in the INDEX.  A bit sloppy since
    # the name could appear as a substring of another port name, or in
    # a non-dependency field.
#    affected = -1
#    for i in file(indexfile):
#        if portname in i:
#            affected += 1

# This counts multiple matches in the same line which is not what we want
#    affmatch = re.findall(portname, "".join(file(indexfile).readlines()), re.MULTILINE)
#    print affmatch
#    affected = len(affmatch)

    affected = len([1 for i in indexdata if portname in i]) - 1
    
    if affected == -1:
        print "%s not in index" % portname
        exit(1)

    line = 0
    portdir = ''
    bdatetime = ''
    maintainer = ''
    broken = "no"

    res = None

    #print "header %s" % time.time()
    while True:
        i = fh.readline()
        line += 1
        if line > header_lines:
            break

        if i.startswith("port directory:"):
            portdir = "/".join(i.split('/')[3:]).rstrip()

        if i.startswith("maintained by"):
            maintainer = i.split()[2].rstrip()

        if i.startswith("build started at"):
            bdatetime = "_".join(i.split()[4:]).rstrip()

    fh.seek(0)
    if not bzipped:
        lines = mm
        
#    if re.search("Trying build of .* even though it is marked BROKEN", lines, re.MULTILINE):
#        broken = "broken"

#    for j in ml:
#        print "re %s " % time.time()
#        res = j.match(lines)
    st = time.time()
    res = domatch(ml, lines)
    ft = time.time()
    print ft - st
#        if res:
#            break

    #print "read %d lines" % line

    # Classify
    if res:
        reason = res[0]
        tag = res[1]
    else:
        reason = '???'
        tag = 'unknown'

    if len(portdir) == 0 or len(maintainer) == 0 or len(bdatetime) == 0:
        reason = 'cluster'
        tag = 'cluster'

    return "|".join((filename, portname, str(affected), str(logsize), portdir, maintainer, reason, tag, broken, bdatetime, errordir))

if __name__ == "__main__":

    #print "Start %s" % time.time()

    if len(argv) == 1:
        print "usage: %s <logdir> [<indexdir>] [<errordir>]" % argv[0]
        exit(1)

    dirname=argv[1]
    if len(argv) > 2:
        indexdir = argv[2]
    else:
        indexdir = "."
    indexdir = os.path.realpath(indexdir)
    
    if len(argv) > 3:
        errordir = argv[3]
    else:
        errordir = "."
    errordir = os.path.realpath(errordir)

    indexfile = os.path.join(indexdir, "INDEX")
    header_lines = 7

    #print "initmatches %s" % time.time()
    ml = initmatches()
    rt = [0 for i in xrange(len(ml))]

    indexdata = file(indexfile).readlines()

#    for f in os.listdir(dirname):
#        if f.endswith(".log") or f.endswith(".log.bz2"):
#            #print f
#            print processfile(f, indexfile, indexdata, errordir, ml)
    print processfile(argv[1], indexfile, indexdata, errordir, ml)

    for (time, num) in sorted([(val, num) for (num, val) in enumerate(rt)]):
        print "%f %3d : %s" % (time, num, ml[num][0])
    
    print sum(rt)
